package trafficSim;

import java.util.ArrayList;
import java.util.List;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameters;
import repast.simphony.query.space.grid.GridCell;
import repast.simphony.query.space.grid.GridCellNgh;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.SpatialMath;
import repast.simphony.space.continuous.ContinuousSpace;
import repast.simphony.space.continuous.NdPoint;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.util.ContextUtils;
import repast.simphony.util.SimUtilities;

public class Car {
	
	private ContinuousSpace<Object> space;
	private Grid<Object> grid;
	private List<GridPoint> instructions;
	private List<GridPoint> leftPoint;
	private int direction;
	private boolean goLeft;
	private static final GridPoint[] directions = {new GridPoint(0, 1), new GridPoint(1, 0), new GridPoint(0, -1), new GridPoint(-1, 0)};
	private int wait;
	private boolean stopped;
	
	private int time;
	
	private GridPoint test;
	
	public Car(ContinuousSpace<Object> space, Grid<Object> grid, GridPoint startPos, double chance) {
		this.space = space;
		this.grid = grid;
		test = new GridPoint(0, 0);
		time = 0;
		wait = 0;
		instructions = new ArrayList<GridPoint>();
		leftPoint = new ArrayList<GridPoint>();
		direction = findInitialDirection(startPos);
		
		
		if(Math.random() < chance) {
			goLeft = true;
			initialDirections(direction);
		} else {
			goLeft = false;
		}
	}
	
	//Metoda pohybu auta
	@ScheduledMethod(start = 1, interval = 1, priority = 0)
	public void drive() {		
		time++;
		GridPoint currentPos = grid.getLocation(this);
		GridPoint nextPosition = new GridPoint(currentPos.getX() + directions[direction].getX(), currentPos.getY() + directions[direction].getY());
		
		if(isCar(nextPosition)) {
			stopped = true;
			wait++;
			return;
		}
		if(isRed(new GridPoint(nextPosition.getX() + directions[direction].getX(), nextPosition.getY() + directions[direction].getY())) && instructions.isEmpty()) {
			return;
		}
		
		if(!instructions.isEmpty()) {
			stopped = false;
			moveByInstruction();
			return;
		}
		
		if(!leftPoint.isEmpty()) {
			stopped = false;
			moveLeft();
			return;
		}
		
		
		
		if(!isCrossroad(nextPosition, grid)) {
			stopped = false;
			grid.moveTo(this, nextPosition.getX(), nextPosition.getY());
			space.moveTo(this, nextPosition.getX(), nextPosition.getY());
		} else {
			if(isGreenNew(nextPosition)) {
				stopped = false;
				crossroadInstructions(nextPosition);
				moveByInstruction();	
			} else {
				if(!isStart(currentPos)) {
					stopped = true;
					wait++;
				}
			}
			
		}
		if(isExit(currentPos)) {
			System.out.println(wait);
			removeCar();
		}
		
	}
	
	public boolean isStopped() {
		return this.stopped == false;
	}
	
	private boolean isRed(GridPoint position) {
		Iterable<Object> objects = grid.getObjectsAt(position.getX(), position.getY());
		for (Object object : objects) {
			if(object.getClass() == TrafficLight.class) {
				TrafficLight tl = (TrafficLight) object;
				if(tl.isRed() || tl.isOrange()) {
					return true;
				}
			}
		}
		return false;
	}
	
	public void move(GridPoint next) {
		NdPoint nt = new NdPoint(next.getX(), next.getY());
		double angle = SpatialMath.calcAngleFor2DMovement(space, space.getLocation(this), nt);
		space.moveByVector(this, 0.5, angle, 0);
		GridPoint currentPt = new GridPoint((int)space.getLocation(this).getX(), (int)space.getLocation(this).getY());
		grid.moveTo(this, currentPt.getX(), currentPt.getY());
	}
	
	
	private List<Integer> allowedDirections(GridPoint currentPos) {
		List<Integer> result = new ArrayList<Integer>();
		int dirRight = (direction + 1) % 4;
		int dirStraight = direction; 
		int dirLeft = Math.floorMod(direction - 1, 4);
		GridPoint testRight = new GridPoint(currentPos.getX() + directions[direction].getX() + directions[dirRight].getX(), currentPos.getY() + directions[direction].getY() + directions[dirRight].getY());
		GridPoint testStraight = new GridPoint(currentPos.getX() + directions[direction].getX() * 3, currentPos.getY() + directions[direction].getY() * 3);
		GridPoint testLeft = new GridPoint(currentPos.getX() + directions[direction].getX() * 2 + directions[dirLeft].getX() * 2, currentPos.getY() + directions[direction].getY() * 2 + directions[dirLeft].getY() * 2);
		if(isRoad(testLeft)) result.add(dirLeft);
		if(isRoad(testStraight)) result.add(dirStraight);
		if(isRoad(testRight)) result.add(dirRight);
		return result;
	}
	
	private void crossroadInstructions(GridPoint currentPos) {
		List<Integer> allowedDirections = allowedDirections(currentPos);
		int newDirection;
		if(goLeft) {
			newDirection = allowedDirections.get(0);
		} else {
			allowedDirections.remove(0);
			SimUtilities.shuffle(allowedDirections, RandomHelper.createUniform());
			newDirection = allowedDirections.get(0);
		}
		
		if(newDirection - direction == -1 || newDirection - direction == 3) {
			//Left
			instructions.add(directions[direction]);
			instructions.add(new GridPoint(directions[direction].getX() + directions[newDirection].getX(), directions[direction].getY() + directions[newDirection].getY()));
			instructions.add(new GridPoint(directions[direction].getX() + directions[newDirection].getX(), directions[direction].getY() + directions[newDirection].getY()));
			instructions.add(directions[newDirection]);
			test = new GridPoint(currentPos.getX(), currentPos.getY());
		} else if(newDirection == direction) {
			//Straight
			instructions.add(directions[newDirection]);
			instructions.add(directions[newDirection]);
			instructions.add(directions[newDirection]);
			instructions.add(directions[newDirection]);
			instructions.add(directions[newDirection]);
		} else if(newDirection - direction == 1 || newDirection - direction == -3) {
			//Right
			instructions.add(directions[direction]);
			instructions.add(directions[newDirection]);
			instructions.add(directions[newDirection]);
		}
		for (GridPoint direction : instructions) {
			test = new GridPoint(test.getX() + direction.getX(), test.getY() + direction.getY());
			
		}
		direction = newDirection;
	}
	
	private void initialDirections(int direction) {
		if(direction == 0) {
			instructions.add(directions[0]);
			instructions.add(directions[3]);
			instructions.add(directions[0]);
		} else if(direction == 1) {
			instructions.add(directions[1]);
			instructions.add(directions[0]);
			instructions.add(directions[1]);
		} else if(direction == 2) {
			instructions.add(directions[2]);
			instructions.add(directions[1]);
			instructions.add(directions[2]);
		} else if(direction == 3) {
			instructions.add(directions[3]);
			instructions.add(directions[2]);
			instructions.add(directions[3]);
		}
	}
	
	private void moveLeft() {
		GridPoint pt = leftPoint.get(0);
		if(!isCar(pt)) {
			if(!pt.equals(grid.getLocation(this))) {
				NdPoint myPoint = space.getLocation(this);
				NdPoint otherPoint = new NdPoint(pt.getX(), pt.getY());
				double angle = SpatialMath.calcAngleFor2DMovement(space, myPoint, otherPoint);
				space.moveByVector(this, 1, angle, 0);
				myPoint = space.getLocation(this);
				grid.moveTo(this, (int)myPoint.getX(), (int)myPoint.getY());
			} else {
				leftPoint.remove(0);
				instructions.add(directions[direction]);
			}
		}
	}
	
	private void moveByInstructionSpace() {
		test = new GridPoint(0,0);
		GridPoint dir = instructions.get(0);
		
		GridPoint currentPos = grid.getLocation(this);
		GridPoint nextPosition = new GridPoint(currentPos.getX() + dir.getX(), currentPos.getY() + dir.getY());
		
		if(!isCar(nextPosition)) {			
			move(nextPosition);
			if(grid.getLocation(this) == nextPosition) {
				instructions.remove(0);
			}
		}		
	}
	
	private void moveByInstruction() {
		test = new GridPoint(0,0);
		GridPoint dir = instructions.get(0);
		
		GridPoint currentPos = grid.getLocation(this);
		GridPoint nextPosition = new GridPoint(currentPos.getX() + dir.getX(), currentPos.getY() + dir.getY());

		
		if(!isCar(nextPosition)) {
			instructions.remove(0);
			
			grid.moveTo(this, nextPosition.getX(), nextPosition.getY());
			space.moveTo(this, nextPosition.getX(), nextPosition.getY());
		}
			
	}
	
	private boolean isCar(GridPoint nextPos) {
		GridCellNgh<Car> nghCreator = new GridCellNgh<Car>(grid, nextPos, Car.class, 0, 0);
		List<GridCell<Car>> gridCells = nghCreator.getNeighborhood(true);
		for (GridCell<Car> gridCell : gridCells) {
			if(gridCell.size() > 0) {
				return true;
			}
		}
		return false;
	}
	
	private boolean isGreenNew(GridPoint currentPos) {
		Iterable<Object> objects = grid.getObjectsAt(currentPos.getX(), currentPos.getY());
		for (Object object : objects) {
			if(object.getClass() == TrafficLight.class) {
				TrafficLight tl = (TrafficLight) object;
				if(tl.isGreen()) {
					return true;
				}
			}
		}
		return false;
	}
	
	private boolean isGreen(GridPoint currentPos) {
		GridCellNgh<Green> nghCreator = new GridCellNgh<Green>(grid, currentPos, Green.class, 0, 0);
		List<GridCell<Green>> gridCells = nghCreator.getNeighborhood(true);
		for (GridCell<Green> gridCell : gridCells) {
			if(gridCell.size() > 0) {
				return true;
			}
		}
		return false;
	}
	
	private boolean isStart(GridPoint currentPos) {
		GridCellNgh<Start> nghCreator = new GridCellNgh<Start>(grid, currentPos, Start.class, 0, 0);
		List<GridCell<Start>> gridCells = nghCreator.getNeighborhood(true);
		for (GridCell<Start> gridCell : gridCells) {
			if(gridCell.size() > 0) {
				return true;
			}
		}
		return false;
	}
	
	private boolean isExit(GridPoint currentPos) {
		GridCellNgh<Exit> nghCreator = new GridCellNgh<Exit>(grid, currentPos, Exit.class, 0, 0);
		List<GridCell<Exit>> gridCells = nghCreator.getNeighborhood(true);
		for (GridCell<Exit> gridCell : gridCells) {
			if(gridCell.size() > 0) {
				return true;
			}
		}
		return false;
	}
	
	private boolean isExitTest(GridPoint currentPos) {
		GridCell<Exit> test = new GridCell<Exit>(currentPos, Exit.class);
		System.out.println(test.getClass());
		return false;
	}
	
	private void drive2(GridPoint pt) {
		NdPoint currentPos = space.getLocation(this);
		NdPoint nextPos = new NdPoint(pt.getX(), pt.getY());
		double angle = SpatialMath.calcAngleFor2DMovement(space, currentPos, nextPos);
		space.moveByVector(this, 0.2, angle, 0);
		NdPoint myPoint = space.getLocation(this); 
		grid.moveTo(this, (int)myPoint.getX(), (int)myPoint.getY());
		
	}
	
	private boolean isRoad(GridPoint pt) {
		boolean result = false;
		GridCellNgh<Road> nghCreator = new GridCellNgh<Road>(grid, pt, Road.class, 0, 0);
		List<GridCell<Road>> gridCells = nghCreator.getNeighborhood(true);
		for (GridCell<Road> gridCell : gridCells) {
			if(gridCell.size() > 0) {
				result = true;
			}
		}
		return result;
	}
	
	private boolean isCrossroad(GridPoint pt, Grid<Object> grid) {
		boolean result = false;
		GridCellNgh<Crossroad> nghCreator = new GridCellNgh<Crossroad>(grid, pt, Crossroad.class, 0, 0);
		List<GridCell<Crossroad>> gridCells = nghCreator.getNeighborhood(true);
		for (GridCell<Crossroad> gridCell : gridCells) {
			GridPoint currentPos = grid.getLocation(this);
			if(gridCell.getPoint().getX() == currentPos.getX() || gridCell.getPoint().getY() == currentPos.getY()) {
				if(gridCell.size() == 1) {
					result = true;
				}
			}
		}
		return result;
	}
	
	private int findInitialDirection(GridPoint startPos) {
		GridPoint pt = null;
		GridCellNgh<Road> nghCreator = new GridCellNgh<Road>(grid, startPos, Road.class, 1, 1);
		List<GridCell<Road>> gridCells = nghCreator.getNeighborhood(true);
		for (GridCell<Road> gridCell : gridCells) {
			if(gridCell.size() == 1 && (gridCell.getPoint().getX() == startPos.getX() || gridCell.getPoint().getY() == startPos.getY())) {
				 pt = gridCell.getPoint();
			}
		}
		GridPoint dir = new GridPoint(pt.getX() - startPos.getX(), pt.getY() - startPos.getY());
		int result = 0;
		if(dir.getY() == 1) {
			result = 0;
		} else if(dir.getX() == 1) {
			result = 1;
		} else if(dir.getY() == -1) {
			result = 2;
		} else if(dir.getX() == -1) {
			result = 3;
		}
		return result;
	}
	
	public void removeCar() {
		repast.simphony.context.Context<Object> context = ContextUtils.getContext(this);
		context.remove(this);
	}

}
