package trafficSim;

import java.util.ArrayList;
import java.util.List;
import java.util.stream.StreamSupport;

import repast.simphony.context.Context;
import repast.simphony.context.space.continuous.ContinuousSpaceFactory;
import repast.simphony.context.space.continuous.ContinuousSpaceFactoryFinder;
import repast.simphony.context.space.grid.GridFactory;
import repast.simphony.context.space.grid.GridFactoryFinder;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.parameter.Parameters;
import repast.simphony.query.space.grid.GridCell;
import repast.simphony.query.space.grid.GridCellNgh;
import repast.simphony.space.continuous.ContinuousSpace;
import repast.simphony.space.continuous.RandomCartesianAdder;
import repast.simphony.space.grid.WrapAroundBorders;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridBuilderParameters;
import repast.simphony.space.grid.GridPoint;
import repast.simphony.space.grid.SimpleGridAdder;

public class RoadBuilder implements ContextBuilder<Object> {
	
	private List<GridPoint> entries = new ArrayList<GridPoint>();
	private List<GridPoint> entries2 = new ArrayList<GridPoint>();
	private List<GridPoint> roads = new ArrayList<GridPoint>();
	private List<GridPoint> exits = new ArrayList<GridPoint>();
	private List<GridPoint> starts = new ArrayList<GridPoint>();
	private List<GridPoint> crossroads = new ArrayList<GridPoint>();
	private List<ExitStart> exitStarts = new ArrayList<ExitStart>();
	
	private Context<Object> context;
	
	private Grid<Object> grid;
	private ContinuousSpace<Object> space;

	@Override
	public Context build(Context<Object> context) {
		
		this.context = context;
		
		Parameters params = RunEnvironment.getInstance().getParameters();
		
		int gridDims = params.getInteger("grid_dimensions");
		boolean isFullyActuated = params.getBoolean("fully_actuated");
		boolean isPartiallyActuated = params.getBoolean("partially_actuated");
		boolean isStatic = params.getBoolean("static");
		double chance = params.getDouble("turn_chance");
		
		context.setId("trafficSim");
		
		ContinuousSpaceFactory spaceFactory = ContinuousSpaceFactoryFinder.createContinuousSpaceFactory(null);
		space = spaceFactory.createContinuousSpace("space", context, new RandomCartesianAdder<Object>(), new repast.simphony.space.continuous.StrictBorders(), gridDims, gridDims);
		GridFactory gridFactory = GridFactoryFinder.createGridFactory(null);
		grid = gridFactory.createGrid("grid", context, new GridBuilderParameters<Object>(new WrapAroundBorders(), new SimpleGridAdder<Object>(), true, gridDims ,gridDims));
		
		int gridWidth = grid.getDimensions().getWidth();
		int gridHeight = grid.getDimensions().getHeight();

		createEntries2(grid);
		
		entries2.addAll(entries);
		
		while(!entries.isEmpty()) {
			GridPoint start = entries.get(0);
			GridPoint end = entries.get(1);
			generateRoads2Lane(start, end, grid, space, context);
			generateExitStarts(start, end, grid, space, context);
			
			entries.remove(start);
			entries.remove(end);
		}
		
		while(!entries2.isEmpty()) {
			GridPoint start = entries2.get(0);
			GridPoint end = entries2.get(1);
			entries2.remove(start);
			entries2.remove(end);
		}
		
		filterEndPointTest(exitStarts, grid, space);

		generateCrossroads(grid, space, context);

		for(int i = 0; i < exitStarts.size(); i++) {
			ExitStart exitStart = exitStarts.get(i);
			GridPoint startCoords = exitStart.getStart();
			GridPoint exit1Coords = exitStart.getExit1();
			GridPoint exit2Coords = exitStart.getExit2();
			Start start = new Start(context, grid, space, startCoords);
			Exit exit1 = new Exit();
			Exit exit2 = new Exit();
			context.add(start);
			context.add(exit1);
			context.add(exit2);
			grid.moveTo(start, startCoords.getX(), startCoords.getY());
			space.moveTo(start, startCoords.getX(), startCoords.getY());
			grid.moveTo(exit1, exit1Coords.getX(), exit1Coords.getY());
			space.moveTo(exit1, exit1Coords.getX(), exit1Coords.getY());
			grid.moveTo(exit2, exit2Coords.getX(), exit2Coords.getY());
			space.moveTo(exit2, exit2Coords.getX(), exit2Coords.getY());
		}
		
		TrafficLightController trafficLight = new TrafficLightController(new GridPoint(gridWidth / 2, gridHeight / 2), context, grid, space, isPartiallyActuated, isFullyActuated, isStatic);
		context.add(trafficLight);
		
		CarSpawner carSpawner = new CarSpawner(context, space, grid, exitStarts, chance);
		context.add(carSpawner);
		
		roads = new ArrayList<GridPoint>();
		exitStarts = new ArrayList<ExitStart>();
		entries = new ArrayList<GridPoint>();
		starts = new ArrayList<GridPoint>();
		exits = new ArrayList<GridPoint>();
		return context;
	}
	
	private void createEntries2(Grid<Object> grid) {
		GridPoint pt1 = new GridPoint(3, grid.getDimensions().getHeight()/2);
		GridPoint pt2 = new GridPoint(grid.getDimensions().getWidth() - 3, grid.getDimensions().getHeight()/2);
		GridPoint pt3 = new GridPoint(grid.getDimensions().getWidth()/2, 3);
		GridPoint pt4 = new GridPoint(grid.getDimensions().getWidth()/2, grid.getDimensions().getHeight() - 3);
		entries.add(pt1);
		entries.add(pt2);
		entries.add(pt3);
		entries.add(pt4);
	}
	
	//Tvorba entries mezi ktermi se nsledn bude tvoit cesty(road)
	private void createEntries(int amount, Grid<Object> grid, int distance) {
		GridPoint pt1 = null;
		GridPoint pt2 = null;
		while(entries.size() < amount * 2) {
			if(Math.random() < 0.5) {
				int randomY1 = (int)(Math.random() * (grid.getDimensions().getHeight() / 2) + 2);
				int testerX = (int)(Math.random() * grid.getDimensions().getWidth());
				int subtractor = testerX < (grid.getDimensions().getWidth() / 2) ? 2 : -2;
				int randomX = testerX + subtractor;
				int randomY2 = grid.getDimensions().getHeight() / 2 + ((int)(Math.random() * (grid.getDimensions().getHeight()) / 2) - 2);
				pt1 = new GridPoint(randomX, randomY1);
				pt2 = new GridPoint(randomX, randomY2);
			} else {
				int randomX1 = (int)(Math.random() * (grid.getDimensions().getWidth()) / 2 + 2);
				int testerY = (int)(Math.random() * grid.getDimensions().getHeight());
				int subtractor = testerY < (grid.getDimensions().getHeight() / 2) ? 2 : -2;
				int randomY = testerY + subtractor;
				int randomX2 = grid.getDimensions().getWidth() / 2 + (int)(Math.random() * (grid.getDimensions().getWidth() / 2) - 2);
				pt1 = new GridPoint(randomX1, randomY);
				pt2 = new GridPoint(randomX2, randomY);
			}
			
			if(!validateAlt(pt1, entries, distance) && !validateAlt(pt2, entries, distance)) {
				entries.add(pt1);
				entries.add(pt2);
			}
		}
	}

	//Tvorba objekt exit+start podle orientace silnice
	private void generateExitStarts(GridPoint startPT, GridPoint endPT, Grid<Object> grid, ContinuousSpace<Object> space, Context<Object> context) {
		if(startPT.getX() == endPT.getX()) {
			GridPoint start1 = new GridPoint(startPT.getX() + 2, startPT.getY() - 1);
			GridPoint exit1 = new GridPoint(startPT.getX(), startPT.getY() - 1);
			GridPoint exit2 = new GridPoint(startPT.getX() - 1, startPT.getY() - 1);
			
			starts.add(start1);
			exits.add(exit1);
			
			
			exitStarts.add(new ExitStart(start1, exit1, exit2));
			
			GridPoint start2 = new GridPoint(endPT.getX() - 1, endPT.getY() + 1);
			GridPoint exit3 = new GridPoint(endPT.getX() + 1, endPT.getY() + 1);
			GridPoint exit4 = new GridPoint(endPT.getX() + 2, endPT.getY() + 1);
			
			starts.add(start2);
			exits.add(exit3);
			exits.add(exit4);
			
			exitStarts.add(new ExitStart(start2, exit3, exit4));
		}
		if(startPT.getY() == endPT.getY()) {
			GridPoint start1 = new GridPoint(startPT.getX() - 1, startPT.getY() - 1);
			GridPoint exit1 = new GridPoint(startPT.getX() - 1, startPT.getY() + 1);
			GridPoint exit2 = new GridPoint(startPT.getX() - 1, startPT.getY() + 2);
			
			starts.add(start1);
			exits.add(exit1);
			exits.add(exit2);
			
			exitStarts.add(new ExitStart(start1, exit1, exit2));
			
			GridPoint start2 = new GridPoint(endPT.getX() + 1, endPT.getY() + 2);
			GridPoint exit3 = new GridPoint(endPT.getX() + 1, endPT.getY());
			GridPoint exit4 = new GridPoint(endPT.getX() + 1, endPT.getY() - 1);
			
			starts.add(start2);
			exits.add(exit3);
			exits.add(exit4);
			
			exitStarts.add(new ExitStart(start2, exit3, exit4));
		}
	}
	
	private boolean isRoad(GridPoint point) {
		Iterable<Object> objects = grid.getObjectsAt(point.getX(), point.getY());
		for (Object object : objects) {
			if(object.getClass() == Road.class) {
				return true;
			}
		}
		return false;
	}
	
	//Generovn silnic(road) mezi dvma entries	
	private void generateRoads2Lane(GridPoint start, GridPoint end, Grid<Object> grid, ContinuousSpace<Object> space, Context<Object> context) {
		if(start.getX() == end.getX()) {
			for(int i = start.getY(); i != end.getY() + 1; i ++) {
				roads.add(new GridPoint(start.getX(), i));
				roads.add(new GridPoint(start.getX(), i + 1));
				roads.add(new GridPoint(start.getX(), i - 1));
				roads.add(new GridPoint(start.getX(), i + 2));
				Road road1 = new Road();
				Road road2 = new Road();
				Road road3 = new Road();
				Road road4 = new Road();
				context.add(road1);
				context.add(road2);
				context.add(road3);
				context.add(road4);
				
				grid.moveTo(road1, start.getX(), i);
				space.moveTo(road1, start.getX(), i);
				grid.moveTo(road2, start.getX() + 1, i);
				space.moveTo(road2, start.getX() + 1, i);
				grid.moveTo(road3, start.getX() - 1, i);
				space.moveTo(road3, start.getX() - 1, i);
				grid.moveTo(road4, start.getX() + 2, i);
				space.moveTo(road4, start.getX() + 2, i);
			}
		}
		if(start.getY() == end.getY()) {
			for(int i = start.getX(); i != end.getX() + 1; i++) {
				roads.add(new GridPoint(i, start.getY()));
				roads.add(new GridPoint(i + 1, start.getY()));
				roads.add(new GridPoint(i - 1, start.getY()));
				roads.add(new GridPoint(i + 2, start.getY()));
				
				Road road1 = new Road();
				Road road2 = new Road();
				Road road3 = new Road();
				Road road4 = new Road();
				
				context.add(road1);
				context.add(road2);
				context.add(road3);
				context.add(road4);
				
				grid.moveTo(road1, i, start.getY());
				space.moveTo(road1, i, start.getY());
				
				grid.moveTo(road2, i , start.getY() + 1);
				space.moveTo(road2, i , start.getY() + 1);
				
				grid.moveTo(road3, i , start.getY() - 1);
				space.moveTo(road3, i , start.getY() - 1);
				
				grid.moveTo(road4, i , start.getY() + 2);
				space.moveTo(road4,i , start.getY() + 2);
			}
		}
	}
	
	private void generateRoads(GridPoint start, GridPoint end, Grid<Object> grid, ContinuousSpace<Object> space, Context<Object> context) {
		if(start.getX() == end.getX()) {
			for(int i = start.getY(); i != end.getY() + 1; i ++) {
				roads.add(new GridPoint(start.getX(), i));
				roads.add(new GridPoint(start.getX(), i + 1));
				Road road = new Road();
				Road road2 = new Road();
				context.add(road);
				context.add(road2);
				grid.moveTo(road, start.getX(), i);
				space.moveTo(road, start.getX(), i);
				grid.moveTo(road2, start.getX() + 1, i);
				space.moveTo(road2, start.getX() + 1, i);
				context.add(road);
				context.add(road2);
			}
		}
		if(start.getY() == end.getY()) {
			for(int i = start.getX(); i != end.getX() + 1; i++) {
				roads.add(new GridPoint(i, start.getY()));
				roads.add(new GridPoint(i + 1, start.getY()));
				
				Road road = new Road();
				Road road2 = new Road();
				
				context.add(road);
				context.add(road2);
				
				grid.moveTo(road, i, start.getY());
				space.moveTo(road, i, start.getY());
				
				grid.moveTo(road2, i , start.getY() + 1);
				space.moveTo(road2,i , start.getY() + 1);
				
				context.add(road);
				context.add(road2);
			}
		}
	}

	//Filtrace z entries pokud le na silnici
	private void filterEndPointTest(List<ExitStart> pts, Grid<Object> grid, ContinuousSpace<Object> space) {
		List<ExitStart> localArray = new ArrayList<ExitStart>();
		for (int i = 0; i < pts.size(); i++) {
			GridPoint pt = pts.get(i).getStart();
			if(StreamSupport.stream(grid.getObjectsAt(pt.getX(), pt.getY()).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
				localArray.add(pts.get(i));
			}
		}
		pts.removeAll(localArray);
	}

	//Validace entry, jestli nezasahuje do prostoru jinho entry
	private boolean validateAlt(GridPoint pt, List<GridPoint> exits, int distance) {
		boolean result = false;
		for (GridPoint exit : exits) {
			if((pt.getX() == exit.getX() && pt.getY() == exit.getY())) {
				return true;
			}
			for(int i = 1; i <= distance; i++) {
				int fx1 = exit.getX() + i;
				int fx2 = exit.getX() - i;
				int fy1 = exit.getY() + i;
				int fy2 = exit.getY() - i;
				if(pt.getX() == fx1 || pt.getX() == fx2 || pt.getY() == fy1 || pt.getY() == fy2) {
					result = true;
					break;
				}
				
			}
		}
		return result;
	}
	
	private void generateCrossroads(Grid<Object> grid, ContinuousSpace<Object> space, Context<Object> context) {
		GridPoint center = new GridPoint(grid.getDimensions().getWidth() / 2, grid.getDimensions().getHeight() /2);
		
		for(int x = -1; x <= 2; x++) {
			for(int y = -1; y <= 2; y++) {
				Crossroad cross = new Crossroad();
				context.add(cross);
				grid.moveTo(cross, center.getX() + x, center.getY() + y);
				space.moveTo(cross, center.getX() + x, center.getY() + y);
				context.add(cross);
			}
		}
		
	}
	
	private boolean checkForRoad(GridPoint pos) {
		GridCellNgh<Road> nghCreator = new GridCellNgh<Road>(grid, pos, Road.class, 0, 0);
		List<GridCell<Road>> gridCells = nghCreator.getNeighborhood(true);
		for (GridCell<Road> gridCell : gridCells) {
			if(gridCell.size() > 0) {
				return true;
			}
		}
		return false;
	}
	
	//Nalezen kiovatek ve vygenerovanch silnicch
	private void findCrossdroads(Grid<Object> grid, ContinuousSpace<Object> space, Context<Object> context) {
		for(int x = 2; x < grid.getDimensions().getWidth() - 2; x++) {
			for(int y = 2; y < grid.getDimensions().getHeight() - 2; y++) {
				int amount = 0;
				GridPoint pt1 = new GridPoint(x - 1, y);
				GridPoint pt2 = new GridPoint(x, y - 1);
				GridPoint pt3 = new GridPoint(x + 2, y);
				GridPoint pt4 = new GridPoint(x, y + 2);
				if(StreamSupport.stream(grid.getObjectsAt(pt1.getX(), pt1.getY()).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
					if(StreamSupport.stream(grid.getObjectsAt(pt1.getX(), pt1.getY() + 1).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
						amount++;
					}
				}
				if(StreamSupport.stream(grid.getObjectsAt(pt2.getX(), pt2.getY()).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
					if(StreamSupport.stream(grid.getObjectsAt(pt2.getX() + 1, pt2.getY()).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
						amount++;
					}
				}
				if(StreamSupport.stream(grid.getObjectsAt(pt3.getX(), pt3.getY()).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
					if(StreamSupport.stream(grid.getObjectsAt(pt3.getX(), pt3.getY() + 1).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
						amount++;
					}
				}
				if(StreamSupport.stream(grid.getObjectsAt(pt4.getX(), pt4.getY()).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
					if(StreamSupport.stream(grid.getObjectsAt(pt4.getX() + 1, pt4.getY()).spliterator(), false).anyMatch(Road -> Road.equals(Road))) {
						amount++;
					}
				}
				if(amount >= 3) {
					for(int i = 0; i < 4; i++) {
						int xP = 0;
						int yP = 0;
						switch (i) {
						case 0:
							xP = 0;
							yP = 0;
							break;
						case 1:
							xP = 1;
							yP = 0;
							break;
						case 2:
							xP = 0;
							yP = 1;
							break;
						case 3:
							xP = 1;
							yP = 1;

						default:
							break;
						}
						crossroads.add(new GridPoint(x + xP, y + yP));
						Crossroad cross = new Crossroad();
						context.add(cross);
						grid.moveTo(cross, x + xP, y + yP);
						space.moveTo(cross, x + xP, y + yP);
						context.add(cross);
					}
					new Semafor(x, y, grid, space, context);
					
				}
			}
		}
	}
}
