package trafficSim;

import java.util.ArrayList;
import java.util.List;

import repast.simphony.context.Context;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.space.continuous.ContinuousSpace;
import repast.simphony.space.grid.Grid;
import repast.simphony.space.grid.GridPoint;

public class TrafficLightController {
	
	private static final double fullInterval = 60;
	private static final double OInterval = 4;
	private static final double RGInterval = (fullInterval - (4 * OInterval)) / 4;
	
	//private int interval = 60;
	private int orangeInterval = 5;
	private int redGreenInterval = 10;
	
	private int allowedDir;
	private Context<Object> context;
	private GridPoint initialPoint;
	private Grid<Object> grid;
	private ContinuousSpace<Object> space;
	
	//Promnn pro semafory (pozice, boolean, erven, zelen, oranov)
	private GridPoint topL, topS, rightL, rightS, bottomL, bottomS, leftL, leftS;	
	private int topLC, topSC, rightLC, rightSC, bottomLC, bottomSC, leftLC, leftSC;
	private TrafficLight topLT, topST, rightLT, rightST, bottomLT, bottomST, leftLT, leftST;
	
	//Senzory
	private Sensor topLSens, topSSens, rightLSens, rightSSens, bottomLSens, bottomSSens, leftLSens, leftSSens;
	
	
	private int sensP;
	private int maxInterval;
	
	private int orangeTime = 6;
	private int orangeTimeSetup = 6;
	
	private int minimumMainTime = 40;
	private int minimumMainTimeBackup = minimumMainTime;
	
	private int maximumMinorTimeBackup = 10;
	private int maximumMinorTime = 10;
	
	private int timePerCar = 6;
	
	
	
	private int state1Time, state2Time, state3Time, state4Time;
	private int maxState1Time, maxState2Time, maxState3Time, maxState4Time;
	
	private boolean state1, state2, state3, state4;
	
	private boolean canChangeToOrange, canChangeToGreen;
	private boolean canChange;
	
	private boolean openMainRoad;
	
	private int addition;
	
	boolean isOrange = false;
	private int minMain = 20;
	
	private boolean isDynamic;
	private boolean isCounting = false;
	
	private int carAmount1, carAmount2;
	
	private int step1Interval, step2Interval, step3Interval, step4Interval;
	private int orange1Interval, orange2Interval, orange3Interval, orange4Interval;
	
	boolean isFullyActuated, isPartiallyActuated, isStatic;
	
	private List<Integer> priorityList = new ArrayList<Integer>();	
	
	//Interval lengths for STATIC COLOR CHANGER
	private int state1Interval, state2Interval, state3Interval, state4Interval;
	
	public TrafficLightController(GridPoint initialPoint, Context<Object> context, Grid<Object> grid, ContinuousSpace<Object> space, boolean isPartiallyActuated, boolean isFullyActuated, boolean isStatic) {
		this.initialPoint = initialPoint;
		this.context = context;
		this.grid = grid;
		this.space = space;
		int x = initialPoint.getX();
		int y = initialPoint.getY();
		
		topL = new GridPoint(x, y + 2);
		topS = new GridPoint(x - 1, y + 2);
		rightL = new GridPoint(x + 2, y + 1);
		rightS = new GridPoint(x + 2, y + 2);
		bottomL = new GridPoint(x + 1, y - 1);
		bottomS = new GridPoint(x + 2, y - 1);
		leftL = new GridPoint(x - 1, y);
		leftS = new GridPoint(x - 1, y - 1);
		
		/*
		topL = new GridPoint(x, y + 3);
		topS = new GridPoint(x - 1, y + 3);
		rightL = new GridPoint(x + 3, y + 1);
		rightS = new GridPoint(x + 3, y + 2);
		bottomL = new GridPoint(x + 1, y - 2);
		bottomS = new GridPoint(x + 2, y - 2);
		leftL = new GridPoint(x - 2, y);
		leftS = new GridPoint(x - 2, y - 1);
		*/
		
		this.isFullyActuated = isFullyActuated;
		this.isPartiallyActuated = isPartiallyActuated;
		this.isStatic = isStatic;
		
		topLC = 0; topSC = 0; rightLC = 0; rightSC = 0; bottomLC = 0; bottomSC = 0; leftLC = 0; leftSC = 0;

		
		//state1Interval = (int) (intervalTime * 0.05 + 1); state2Interval = (int) (intervalTime * 0.05 + 1); state3Interval = (int) (intervalTime * 0.8 + 1); state4Interval = (int)(intervalTime * 0.1 + 1);
		state1Interval = 8; state2Interval = 8; state3Interval = 50; state4Interval = 8;
		
		state1 = false; state2 = false; state3 = false; state4 = false;
		
		state1Time = 0; state2Time = 0; state3Time = 0; state4Time = 0;
		maxState1Time = 0; maxState2Time = 0; maxState3Time = 0; maxState4Time = 0;
		
		step1Interval = redGreenInterval;
		sensP = 5;
		maxInterval = sensP * 3;
		addition = 2;
		openMainRoad = true;

		generateLightsNew();
		initialize();
		setupSensors();
	}
	
	public boolean isState1() {
		return this.state1;
	}
	
	public boolean isState2() {
		return this.state2;
	}
	
	public boolean isState3() {
		return this.state3;
	}
	
	public boolean isState4() {
		return this.state4;
	}
	
	@ScheduledMethod(start = 1, interval = 1)
	public void colorChanger() {
		
		if(isFullyActuated) {
			colorChangerFullyActuated();
		}
		if(isPartiallyActuated) {
			colorChangerPartiallyActuatedTESTER();
		}
		if(isStatic) {
			colorChangerStatic();
		}
	}

	//ACTUATED COLOR CHANGERS__________________________________________________________________________________________________________________________________________________________
	
	
	@ScheduledMethod(start = 1, interval = 1)
	public void detectorTester() {
		//topL = state2
		if(topLSens.getAmountOfCars() > 0) {
			topLC++;
			if(maxState2Time < maximumMinorTime) {
				maxState2Time = calculateTimes(topLC, bottomLC);
				state2Time = calculateTimes(topLC, bottomLC);
			}
			priorityAdder(2);
			topLSens.setAmountOfCars(0);
		}
		//topS = state1
		if(topSSens.getAmountOfCars() > 0) {
			topSC++;
			if(maxState1Time < maximumMinorTime) {
				maxState1Time = calculateTimes(topSC, bottomSC);
				state1Time = calculateTimes(topSC, bottomSC);
			}
			priorityAdder(1);
			topSSens.setAmountOfCars(0);
		}
		//rightL = state4
		if(rightLSens.getAmountOfCars() > 0) {
			rightLC++;
			if(maxState4Time < maximumMinorTime) {
				maxState4Time = calculateTimes(rightLC, leftLC);
				state4Time = calculateTimes(rightLC, leftLC);
			}
			priorityAdder(4);
			rightLSens.setAmountOfCars(0);
		}
		//rightS - Main State = state3
		if(isFullyActuated) {
			if(rightSSens.getAmountOfCars() > 0) {
				rightSC++;
				if(maxState3Time < maximumMinorTime) {
					maxState3Time = calculateTimes(rightSC, leftSC);
					state3Time = calculateTimes(rightSC, leftSC);
				}
				priorityAdder(3);
				rightSSens.setAmountOfCars(0);
			}
		}
		//bottomL = state2
		if(bottomLSens.getAmountOfCars() > 0) {
			bottomLC++;
			if(maxState2Time < maximumMinorTime) {
				maxState2Time = calculateTimes(bottomLC, topLC);
				state2Time = calculateTimes(bottomLC, topLC);
			}
			priorityAdder(2);
			bottomLSens.setAmountOfCars(0);
		}
		//bottomS = state1
		if(bottomSSens.getAmountOfCars() > 0) {
			bottomSC++;
			if(maxState1Time < maximumMinorTime) {
				maxState1Time = calculateTimes(bottomSC, topSC);
				state1Time = calculateTimes(bottomSC, topSC);
			}
			priorityAdder(1);
			bottomSSens.setAmountOfCars(0);
		}
		//leftL = state4
		if(leftLSens.getAmountOfCars() > 0) {
			leftLC++;
			if(maxState4Time < maximumMinorTime) {
				maxState4Time = calculateTimes(leftLC, rightLC);
				state4Time = calculateTimes(leftLC, rightLC);
			}
			priorityAdder(4);
			leftLSens.setAmountOfCars(0);
		}
		//leftS - Main State = state3
		if(isFullyActuated) {
			if(leftSSens.getAmountOfCars() > 0) {
				leftSC++;
				if(maxState3Time < maximumMinorTime) {
					maxState3Time = calculateTimes(leftSC, rightSC);
					state3Time = calculateTimes(leftSC, rightSC);
				}
				priorityAdder(3);
				leftSSens.setAmountOfCars(0);
			}
		}
	}
	
	public void detector() {
		//topL = state2
		if(topLSens.getAmountOfCars() > 0) {
			topLC++;
			if(maxState2Time < maximumMinorTime) {
				maxState2Time += timePerCar;
				state2Time += timePerCar;
			} else {
				state2Time = maximumMinorTime;
			}
			priorityAdder(2);
			topLSens.setAmountOfCars(0);
		}
		//topS = state1
		if(topSSens.getAmountOfCars() > 0) {
			topSC++;
			if(maxState1Time < maximumMinorTime) {
				maxState1Time += timePerCar;
				state1Time += timePerCar;
			} else {
				state1Time = maximumMinorTime;
			}
			priorityAdder(1);
			topSSens.setAmountOfCars(0);
		}
		//rightL = state4
		if(rightLSens.getAmountOfCars() > 0) {
			rightLC++;
			if(maxState4Time < maximumMinorTime) {
				maxState4Time += timePerCar;
				state4Time += timePerCar;
			} else {
				state4Time = maximumMinorTime;
			}
			priorityAdder(4);
			rightLSens.setAmountOfCars(0);
		}
		//rightS - Main State = state3
		if(isFullyActuated) {
			if(rightSSens.getAmountOfCars() > 0) {
				rightSC++;
				if(maxState3Time < maximumMinorTime) {
					maxState3Time += timePerCar;
					state3Time += timePerCar;
				} else {
					state3Time = maximumMinorTime;
				}
				priorityAdder(3);
				rightSSens.setAmountOfCars(0);
			}
		}
		//bottomL = state2
		if(bottomLSens.getAmountOfCars() > 0) {
			bottomLC++;
			if(maxState2Time < maximumMinorTime) {
				maxState2Time += timePerCar;
				state2Time += timePerCar;
			} else {
				state2Time = maximumMinorTime;
			}
			priorityAdder(2);
			bottomLSens.setAmountOfCars(0);
		}
		//bottomS = state1
		if(bottomSSens.getAmountOfCars() > 0) {
			bottomSC++;
			if(maxState1Time < maximumMinorTime) {
				maxState1Time += timePerCar;
				state1Time += timePerCar;
			} else {
				state1Time = maximumMinorTime;
			}
			priorityAdder(1);
			bottomSSens.setAmountOfCars(0);
		}
		//leftL = state4
		if(leftLSens.getAmountOfCars() > 0) {
			leftLC++;
			if(maxState4Time < maximumMinorTime) {
				maxState4Time += timePerCar;
				state4Time += timePerCar;
			} else {
				state4Time = maximumMinorTime;
			}
			priorityAdder(4);
			leftLSens.setAmountOfCars(0);
		}
		//leftS - Main State = state3
		if(isFullyActuated) {
			if(leftSSens.getAmountOfCars() > 0) {
				leftSC++;
				if(maxState3Time < maximumMinorTime) {
					maxState3Time += timePerCar;
					state3Time += timePerCar;
				} else {
					state3Time = maximumMinorTime;
				}
				priorityAdder(3);
				leftSSens.setAmountOfCars(0);
			}
		}
	}
	
	public void priorityAdder(int priority) {
		if(!priorityList.contains(priority)) priorityList.add(priority);
	}
	
	public void timeIncreaser(int maxStateTime, int stateTime, int carCounter, int state) {
		carCounter++;
		if(maxStateTime < maximumMinorTime) {
			maxStateTime += timePerCar;
			stateTime += timePerCar;
		} else {
			stateTime = maximumMinorTime;
		}
		
		if(isPartiallyActuated) {
			if(!priorityList.contains(state)) priorityList.add(state);
		} else if(isFullyActuated) {
			priorityList.add(state);
		}
	}
	
	private int calculateTimes(int state1CarAmount, int state2CarAmount) {
		int openTime = 0;
		if(state1CarAmount > state2CarAmount) {
			openTime = state1CarAmount * timePerCar;
		} else {
			openTime = state2CarAmount * timePerCar;
		}
		return openTime;
	}
	
	//Change to state1 for actuated controllers
	private void changeToState1() {
		canChange = false;
		state1 = true;
		if(orangeTime > 0) {
			if(orangeTime == orangeTimeSetup) {
				setOrangeNew();
				//setOrange();
			}
			orangeTime--;
			return;
		}
		if(state1Time > 0) {
			setState1New();
			state1Time--;
		}
		if(state1Time == 0) {
			topSC = 0;
			bottomSC = 0;
			priorityList.remove(0);
			orangeTime = orangeTimeSetup;
			maxState1Time = 0;
			openMainRoad = true;
			state1 = false;
		}
	}
	
	//Change to state2 for actuated controllers
	private void changeToState2() {
		canChange = false;
		state2 = true;
		if(orangeTime > 0) {
			if(orangeTime == orangeTimeSetup) {
				setOrangeNew();
			}
			orangeTime--;
			return;
		}
		if(state2Time > 0) {
			setState2New();
			state2Time--;
		}
		if(state2Time == 0) {
			topLC = 0;
			bottomLC = 0;
			priorityList.remove(0);
			orangeTime = orangeTimeSetup;
			maxState2Time = 0;
			openMainRoad = true;
			state2 = false;
		}
	}
	
	//Change to state3 for FULLY ACTUATED CONTROLLER
	private void changeToState3FA() {
		canChange = false;
		state3 = true;
		if(orangeTime > 0) {
			if(orangeTime == orangeTimeSetup) {
				setOrangeNew();
				//setOrange();
			}
			orangeTime--;
			return;
		}
		if(state3Time > 0) {
			setState3New();
			state3Time--;
		}
		if(state3Time == 0) {
			leftSC = 0;
			rightSC = 0;
			priorityList.remove(0);
			orangeTime = orangeTimeSetup;
			maxState3Time = 0;
			openMainRoad = true;
			state3 = false;
		}
	}
	
	//MAIN STATE - Change to state3 for PARTIALLY ACTUATED CONTROLLER
	private void changeToState3PA() {
		canChange = false;
		if(!state3) {
			if(orangeTime > 0) {
				if(orangeTime == orangeTimeSetup) {
					setOrangeNew();
				}
				orangeTime--;
				return;
			}
		}
		
		state3 = true;
		setState3New();
		if(minimumMainTime == 1) {
			state3 = false;
		}
	}
	
	//Change to state4 for actuated controllers
	private void changeToState4() {
		state4 = true;
		if(orangeTime > 0) {
			if(orangeTime == orangeTimeSetup) {
				setOrangeNew();
			}
			orangeTime--;
			return;
		}
		if(state4Time > 0) {
			setState4New();
			state4Time--;
		}
		if(state4Time == 0) {
			leftLC = 0;
			rightLC = 0;
			priorityList.remove(0);
			orangeTime = orangeTimeSetup;
			maxState4Time = 0;
			openMainRoad = true;
			state4 = false;
		}
	}
	
	public void colorChangerFullyActuated() {	
			if(priorityList.size() > 0) {
				int priority = priorityList.get(0);
				if(priority == 1) {
					if(state1Time > 0) {
						changeToState1();
						if(state1Time == 1) {
							changeToState1();
						}
					}
					
				}
				
				if(priority == 2) {
					if(state2Time > 0) {
						changeToState2();
						if(state2Time == 1) {
							changeToState2();
							openMainRoad = true;
							minimumMainTime = minimumMainTimeBackup;
						}
					}
					
				}
				if(priority == 3) {
					if(state3Time > 0) {
						changeToState3FA();
						if(state3Time == 1) {
							changeToState3FA();
							openMainRoad = true;
							minimumMainTime = minimumMainTimeBackup;
						}
					}
				}
				
				//if(!state1 || !state3) {
				if(priority == 4) {
					if(state4Time > 0) {
						changeToState4();
						if(state4Time == 1) {
							changeToState4();
							openMainRoad = true;
							minimumMainTime = minimumMainTimeBackup;
						}
					}
					
				}
			}
			
	}
	
	//TESTING CLASS
	public void colorChangerPartiallyActuatedTESTER() {
		//Main road
		if(priorityList.size() > 0) {
			if(minimumMainTime != 0) {
				minimumMainTime--;
			}
		}
		if(openMainRoad) {
			changeToState3PA();
			if(minimumMainTime > 0) {
				return;
			}
			
			orangeTime = orangeTimeSetup;
		}
		
		openMainRoad = false;
		if(!openMainRoad) {
			int priority = priorityList.get(0);
			if(priority == 1) {
				if(state1Time > 0) {
					changeToState1();
					if(state1Time == 1) {
						changeToState1();
						openMainRoad = true;
						minimumMainTime = minimumMainTimeBackup;
					}
				}
				
			}

			if(priority == 2) {
				if(state2Time > 0) {
					changeToState2();
					if(state2Time == 0) {
						changeToState2();
						openMainRoad = true;
						minimumMainTime = minimumMainTimeBackup;
					}
				}
				
			}

			if(priority == 4) {
				if(state4Time > 0) {
					changeToState4();
					if(state4Time == 1) {
						changeToState4();
						openMainRoad = true;
						minimumMainTime = minimumMainTimeBackup;
					}
				}
				
			}
			
		}	
	}
	
	//COLOR CHANGER - PARTIALLY ACTUATED
	public void colorChangerPartiallyActuated() {
		if(priorityList.size() > 0) {
			if(minimumMainTime != 0) {
				minimumMainTime--;
			}
		}
		if(openMainRoad) {
			changeToState3PA();
			if(minimumMainTime > 0) {
				return;
			}
			orangeTime = orangeTimeSetup;
		}
		openMainRoad = false;
		if(!openMainRoad) {
			int priority = priorityList.get(0);
			if(priority == 1) {
				if(state1Time > 0) {
					changeToState1();
					if(state1Time == 1) {
						changeToState1();
						openMainRoad = true;
						minimumMainTime = minimumMainTimeBackup;
					}
				}
				
			}

			if(priority == 2) {
				if(state2Time > 0) {
					changeToState2();
					if(state2Time == 0) {
						changeToState2();
						openMainRoad = true;
						minimumMainTime = minimumMainTimeBackup;
					}
				}
				
			}

			if(priority == 4) {
				if(state4Time > 0) {
					changeToState4();
					if(state4Time == 1) {
						changeToState4();
						openMainRoad = true;
						minimumMainTime = minimumMainTimeBackup;
					}
				}
				
			}
			
		}	
	}
	
	//COLOR CHANGER STATIC INTERVALS__________________________________________________________________________________________________________________________________________________________
	public void colorChangerStatic() {
		//Step 1 Interval
				if(step1Interval >= 0) {
					if(step1Interval == state1Interval) {
						setState1New();
					}
					step1Interval--;
					if(step1Interval == 0){
						orange1Interval = orangeInterval;
					}
				}
				
				//Step 1 to 2 Orange Interval
				if(orange1Interval >= 0) {
					if(orange1Interval == orangeInterval) {
						setOrangeNew();
					}
					orange1Interval--;
					if(orange1Interval == 0) {
						//step2Interval = redGreenInterval;
						step2Interval = state2Interval;	
					}
				}
				
				//Step 2 Interval
				if(step2Interval >= 0) {
					if(step2Interval == state2Interval) {
						setState2New();
					}
					step2Interval--;
					if(step2Interval == 0) {
						orange2Interval = orangeInterval;
					}
				}
				
				//Step 2 to 3 Orange Interval
				if(orange2Interval >= 0) {
					if(orange2Interval == orangeInterval) {
						setOrangeNew();
					}
					orange2Interval--;
					if(orange2Interval == 0) {
						step3Interval = state3Interval;
					}
				}
				
				//Step 3 Interval
				if(step3Interval >= 0) {
					if(step3Interval == state3Interval) {
						setState3New();
					}
					step3Interval--;
					if(step3Interval == 0) {
						orange3Interval = orangeInterval;
					}
				}
				
				//Step 3 to 4 Orange Interval
				if(orange3Interval >= 0) {
					if(orange3Interval == orangeInterval) {
						setOrangeNew();
					}
					orange3Interval--;
					if(orange3Interval == 0) {
						step4Interval = state4Interval;
					}
				}
				
				//Step 4 Interval
				if(step4Interval >= 0) {
					if(step4Interval == state4Interval) {
						setState4New();
					}
					step4Interval--;
					if(step4Interval == 0) {
						orange4Interval = orangeInterval;
					}
				}
				
				//Step 4 to 1 Orange Interval
				if(orange4Interval >= 0) {
					if(orange4Interval == orangeInterval) {
						setOrangeNew();
					}
					orange4Interval--;
					if(orange4Interval == 0) {
						step1Interval = state1Interval;
					}
				}
	}
	
	public void colorChangerDynamic() {
		//System.out.println(rightSSens.getAmountOfCars());
		//Step 1 Interval
		if(step1Interval >= 0) {
			if(step1Interval == redGreenInterval) {
				setState1New();
			}
			step1Interval--;
			if(step1Interval == 0){
				orange1Interval = orangeInterval;
				carAmount1 = topLSens.getAmountOfCars();
				carAmount2 = bottomLSens.getAmountOfCars();
				leftSSens.setAmountOfCars(0);
				rightSSens.setAmountOfCars(0);
			}
		}
		
		//Step 1 to 2 Orange Interval
		if(orange1Interval >= 0) {
			if(orange1Interval == orangeInterval) {
				setOrangeNew();
			}
			orange1Interval--;
			if(orange1Interval == 0) {
				//step2Interval = redGreenInterval;
				if(topLSens.getAmountOfCars() < bottomLSens.getAmountOfCars()) {
					step2Interval = carAmount2 * 2 + addition;
					
					if(step2Interval > maxInterval) {
						step2Interval = maxInterval;
					}
					redGreenInterval = step2Interval;
				} else {
					step2Interval = carAmount1 * 2 + addition;
					if(step2Interval > maxInterval) {
						step2Interval = maxInterval;
					}
					redGreenInterval = step2Interval;
				}	
			}
		}
		
		//Step 2 Interval
		if(step2Interval >= 0) {
			if(step2Interval == redGreenInterval) {
				setState2New();
			}
			step2Interval--;
			if(step2Interval == 0) {
				orange2Interval = orangeInterval;
				carAmount1 = rightSSens.getAmountOfCars();
				carAmount2 = leftSSens.getAmountOfCars();
				topLSens.setAmountOfCars(0);
				bottomLSens.setAmountOfCars(0);
			}
		}
		
		//Step 2 to 3 Orange Interval
		if(orange2Interval >= 0) {
			if(orange2Interval == orangeInterval) {
				setOrangeNew();
			}
			orange2Interval--;
			if(orange2Interval == 0) {
				if(rightSSens.getAmountOfCars() < leftSSens.getAmountOfCars()) {
					step3Interval = carAmount2 * 2 + addition;
					
					if(step3Interval > maxInterval) {
						step3Interval = maxInterval;
					}
					redGreenInterval = step3Interval;
				} else {
					step3Interval = carAmount1 * 2 + addition;
					if(step3Interval > maxInterval) {
						step3Interval = maxInterval;
					}
					redGreenInterval = step3Interval;
				}
			}
		}
		
		//Step 3 Interval
		if(step3Interval >= 0) {
			if(step3Interval == redGreenInterval) {
				setState3New();
			}
			step3Interval--;
			if(step3Interval == 0) {
				orange3Interval = orangeInterval;
				carAmount1 = rightLSens.getAmountOfCars();
				carAmount2 = leftLSens.getAmountOfCars();
				rightLSens.setAmountOfCars(0);
				leftLSens.setAmountOfCars(0);
			}
		}
		
		//Step 3 to 4 Orange Interval
		if(orange3Interval >= 0) {
			if(orange3Interval == orangeInterval) {
				setOrangeNew();
			}
			orange3Interval--;
			if(orange3Interval == 0) {
				//step4Interval = redGreenInterval;
				if(leftLSens.getAmountOfCars() < rightLSens.getAmountOfCars()) {
					step4Interval = carAmount1 * 2 + addition;
					
					if(step4Interval > maxInterval) {
						step4Interval = maxInterval;
					}
					redGreenInterval = step4Interval;
				} else {
					step4Interval = carAmount2 * 2 + addition;
					if(step4Interval > maxInterval) {
						step4Interval = maxInterval;
					}
					redGreenInterval = step4Interval;
				}
			}
		}
		
		//Step 4 Interval
		if(step4Interval >= 0) {
			if(step4Interval == redGreenInterval) {
				setState4New();
			}
			step4Interval--;
			if(step4Interval == 0) {
				orange4Interval = orangeInterval;
				carAmount1 = bottomSSens.getAmountOfCars();
				carAmount2 = topSSens.getAmountOfCars();
				bottomSSens.setAmountOfCars(0);
				topSSens.setAmountOfCars(0);
			}
		}
		
		//Step 4 to 1 Orange Interval
		if(orange4Interval >= 0) {
			if(orange4Interval == orangeInterval) {
				setOrangeNew();
			}
			orange4Interval--;
			if(orange4Interval == 0) {
				if(topSSens.getAmountOfCars() < bottomSSens.getAmountOfCars()) {
					step1Interval = carAmount1 * 2 + addition;

					if(step1Interval > maxInterval) {
						step1Interval = maxInterval;
					}
					redGreenInterval = step1Interval;
				} else {
					step1Interval = carAmount2 * 2 + addition;
					if(step1Interval > maxInterval) {
						step1Interval = maxInterval;
					}
					redGreenInterval = step1Interval;
				}
			}
		}
	}
	
	/*
	@ScheduledMethod(start = 1, interval = 1)
	public void crossroadDirO() {
		stepCounter++;
		allowedDir = (stepCounter / (openInterval + closedInterval)) % 4;
		if(stepCounter % (openInterval + closedInterval) < closedInterval) allowedDir = -1;
	}
	*/
	
	
	public void setupSensors() {	
		topLSens = new Sensor(grid, space, new GridPoint(topL.getX(), topL.getY() + sensP));
		context.add(topLSens);
		grid.moveTo(topLSens, topL.getX(), topL.getY() + sensP);
		space.moveTo(topLSens, topL.getX(), topL.getY() + sensP);
		
		topSSens = new Sensor(grid, space, new GridPoint(topS.getX(), topS.getY() + sensP));
		context.add(topSSens);
		grid.moveTo(topSSens, topS.getX(), topS.getY() + sensP);
		space.moveTo(topSSens, topS.getX(), topS.getY() + sensP);
		
		rightLSens = new Sensor(grid, space, new GridPoint(rightL.getX() + sensP, rightL.getY()));
		context.add(rightLSens);
		grid.moveTo(rightLSens, rightL.getX() + sensP, rightL.getY());
		space.moveTo(rightLSens, rightL.getX() + sensP, rightL.getY());
		
		rightSSens = new Sensor(grid, space, new GridPoint(rightS.getX() + sensP, rightS.getY()));
		context.add(rightSSens);
		grid.moveTo(rightSSens, rightS.getX() + sensP, rightS.getY());
		space.moveTo(rightSSens, rightS.getX() + sensP, rightS.getY());
		
		bottomLSens = new Sensor(grid, space, new GridPoint(bottomL.getX(), bottomL.getY() - sensP));
		context.add(bottomLSens);
		grid.moveTo(bottomLSens, bottomL.getX(), bottomL.getY() - sensP);
		space.moveTo(bottomLSens, bottomL.getX(), bottomL.getY() - sensP);
		
		bottomSSens = new Sensor(grid, space, new GridPoint(bottomS.getX(), bottomS.getY() - sensP));
		context.add(bottomSSens);
		grid.moveTo(bottomSSens, bottomS.getX(), bottomS.getY() - sensP);
		space.moveTo(bottomSSens, bottomS.getX(), bottomS.getY() - sensP);
		
		leftLSens = new Sensor(grid, space, new GridPoint(leftL.getX() - sensP, leftL.getY()));
		context.add(leftLSens);
		grid.moveTo(leftLSens, leftL.getX() - sensP, leftL.getY());
		space.moveTo(leftLSens, leftL.getX() - sensP, leftL.getY());
		
		leftSSens = new Sensor(grid, space, new GridPoint(leftS.getX() - sensP, leftS.getY()));
		context.add(leftSSens);
		grid.moveTo(leftSSens, leftS.getX() - sensP, leftS.getY());
		space.moveTo(leftSSens, leftS.getX() - sensP, leftS.getY());
	}

	public int getAllowedDir() {
		return allowedDir;
	}
	
	public void generateLightsNew() {
		//topL
		topLT = new TrafficLight();
		move(topLT, topL);
		//topS
		topST = new TrafficLight();
		move(topST, topS);
		//rightL
		rightLT = new TrafficLight();
		move(rightLT, rightL);
		//rightS
		rightST = new TrafficLight();
		move(rightST, rightS);
		//bottomL
		bottomLT = new TrafficLight();
		move(bottomLT, bottomL);
		//bottomS
		bottomST = new TrafficLight();
		move(bottomST, bottomS);
		//leftL
		leftLT = new TrafficLight();
		move(leftLT, leftL);
		//leftS
		leftST = new TrafficLight();
		move(leftST, leftS);
	}
	
	public void move(Object obj, GridPoint pt) {
		context.add(obj);
		grid.moveTo(obj, pt.getX(), pt.getY());
		space.moveTo(obj, pt.getX(), pt.getY());
	}
	
	public void generateLights() {
		Red red1 = new Red();
		Red red2 = new Red();
		Red red3 = new Red();
		Red red4 = new Red();
		Red red5 = new Red();
		Red red6 = new Red();
		Red red7 = new Red();
		Red red8 = new Red();
		
		context.add(red1);
		context.add(red2);
		context.add(red3);
		context.add(red4);
		context.add(red5);
		context.add(red6);
		context.add(red7);
		context.add(red8);
		
		grid.moveTo(red1, topL.getX(), topL.getY());
		space.moveTo(red1, topL.getX(), topL.getY());
		
		grid.moveTo(red2, topS.getX(), topS.getY());
		space.moveTo(red2, topS.getX(), topS.getY());
		
		grid.moveTo(red3, rightL.getX(), rightL.getY());
		space.moveTo(red3, rightL.getX(), rightL.getY());
		
		grid.moveTo(red4, rightS.getX(), rightS.getY());
		space.moveTo(red4, rightS.getX(), rightS.getY());
		
		grid.moveTo(red5, bottomL.getX(), bottomL.getY());
		space.moveTo(red5, bottomL.getX(), bottomL.getY());
		
		grid.moveTo(red6, bottomS.getX(), bottomS.getY());
		space.moveTo(red6, bottomS.getX(), bottomS.getY());
		
		grid.moveTo(red7, leftL.getX(), leftL.getY());
		space.moveTo(red7, leftL.getX(), leftL.getY());
		
		grid.moveTo(red8, leftS.getX(), leftS.getY());
		space.moveTo(red8, leftS.getX(), leftS.getY());
		
		context.add(red1);
		context.add(red2);
		context.add(red3);
		context.add(red4);
		context.add(red5);
		context.add(red6);
		context.add(red7);
		context.add(red8);
	}
	
	private void initialize() {
		
		setState3New();
		//setState3();
		/*
		topLB = false;
		addToContextAndMove(topLR, topL);
		
		topSB = false;
		addToContextAndMove(topSO, topS);
		
		rightLB = false;
		addToContextAndMove(rightLO, rightL);

		rightSB = false;
		addToContextAndMove(rightSR, rightS);
		
		bottomLB = false;
		addToContextAndMove(bottomLR, bottomL);
		
		bottomSB = false;
		addToContextAndMove(bottomSO, bottomS);
		
		leftLB = false;
		addToContextAndMove(leftLO, leftL);
		
		leftSB = false;
		addToContextAndMove(leftSR, leftS);
		*/
	}
	
	
	
	public void setState1New() {
		state1 = true; state2 = false; state3 = false; state4 = false;
		
		//topL
		topLT.setRed();
		//topS
		topST.setGreen();
		//rightL
		rightLT.setRed();
		//rightS
		rightST.setRed();
		//bottomL
		bottomLT.setRed();
		//bottomS
		bottomST.setGreen();
		//leftL
		leftLT.setRed();
		//leftS
		leftST.setRed();
	}
	
	public void setState2New() {
		state1 = false; state2 = true; state3 = false; state4 = false;
		
		//topL
		topLT.setGreen();
		//topS
		topST.setRed();
		//rightL
		rightLT.setRed();
		//rightS
		rightST.setRed();
		//bottomL
		bottomLT.setGreen();
		//bottomS
		bottomST.setRed();
		//leftL
		leftLT.setRed();
		//leftS
		leftST.setRed();
	}
	
	public void setState3New() {
		state1 = false; state2 = false; state3 = true; state4 = false;
		
		//topL
		topLT.setRed();
		//topS
		topST.setRed();
		//rightL
		rightLT.setRed();
		//rightS
		rightST.setGreen();
		//bottomL
		bottomLT.setRed();
		//bottomS
		bottomST.setRed();
		//leftL
		leftLT.setRed();
		//leftS
		leftST.setGreen();
	}
	
	public void setState4New() {
		state1 = false;
		state2 = false;
		state3 = false;
		state4 = true;
		//topL
		topLT.setRed();
		//topS
		topST.setRed();
		//rightL
		rightLT.setGreen();
		//rightS
		rightST.setRed();
		//bottomL
		bottomLT.setRed();
		//bottomS
		bottomST.setRed();
		//leftL
		leftLT.setGreen();
		//leftS
		leftST.setRed();		
	}
	
	public void setOrangeNew() {
		state1 = false;
		state2 = false;
		state3 = false;
		state4 = false;
		//topL
		if(topLT.isGreen()) topLT.setOrange();
		//topS
		if(topST.isGreen()) topST.setOrange();
		//rightL
		if(rightLT.isGreen()) rightLT.setOrange();
		//rightS
		if(rightST.isGreen()) rightST.setOrange();
		//bottomL
		if(bottomLT.isGreen()) bottomLT.setOrange();
		//bottomS
		if(bottomST.isGreen()) bottomST.setOrange();
		//leftL
		if(leftLT.isGreen()) leftLT.setOrange();
		//leftS
		if(leftST.isGreen()) leftST.setOrange();
	}
	
	private void moveObject(Object object, GridPoint pt) {
		grid.moveTo(object, pt.getX(), pt.getY());
		space.moveTo(object, pt.getX(), pt.getY());
		
	}
	
	private void change(Object toRemove1, Object toRemove2, Object toAdd, GridPoint loc) {
		context.remove(toRemove1);
		context.remove(toRemove2);
		context.add(toAdd);
		grid.moveTo(toAdd, loc.getX(), loc.getY());
		space.moveTo(toAdd, loc.getX(), loc.getY());
	}
	
	private void exchange(Object object1, Object object2, GridPoint pt) {
		context.remove(object1);
		context.add(object2);
		grid.moveTo(object2, pt.getX(), pt.getY());
		space.moveTo(object2, pt.getX(), pt.getY());
	}
	
	private void addToContextAndMove(Object object, GridPoint pt) {
		context.add(object);
		grid.moveTo(object, pt.getX(), pt.getY());
		space.moveTo(object, pt.getX(), pt.getY());
	}

	
	public void carDetection() {
		//topL
		topLSens.addCar();
		//topS
		topSSens.addCar();
		//rightL
		rightLSens.addCar();
		//rightS
		rightSSens.addCar();
		//bottomL
		bottomLSens.addCar();
		//bottomS
		bottomSSens.addCar();
		//leftL
		leftLSens.addCar();
		//leftS
		leftSSens.addCar();
	}
}
